
package Spec;

import java.util.*;
import br.ufpe.abaco.compiler.*;

public class MapImplementation {


    CompiledSpecification spec;
    StandardOperators stan;

    MapImplementation(CompiledSpecification s) {
	this.spec = s;
	stan = new StandardOperators(s);

	spec.insertNativeType("map", MapTerm.class);

	spec.insertNativeImplementation("map of _ to _",  new NativeImplementation() {
		public Term evaluate(Term args[]) {
		  if (stan.isData(args[0]) && stan.isData(args[1])) {
		    HashMap nMap = new HashMap();
		    nMap.put(args[0],args[1]);
		    return new MapTerm(nMap,stan);
		  }
		  return null;
		}
	    });
	
	spec.insertNativeImplementation("empty-map", new NativeImplementation() {
		MapTerm emptyMap = new MapTerm(new HashMap(),stan);
		public Term evaluate(Term args[]) {
		    return emptyMap;
		}
	    });
	
	
	spec.insertNativeImplementation("_ at _", new NativeImplementation() {
		public Term evaluate(Term args[]) {
		    if (args[0] instanceof MapTerm) {
			return ((MapTerm)args[0]).at(args[1]);
		    }
		    return null;
		}
	    });
	
	spec.insertNativeImplementation("overlay _ _", new NativeImplementation() {
		public Term evaluate(Term args[]) {
		    if (args[0] instanceof MapTerm && args[1] instanceof MapTerm) {
			return ((MapTerm)args[0]).overlay((MapTerm)args[1]);
		    }
		    return null;
		}
	    });
	
	spec.insertNativeImplementation("disjoint union _ _", new NativeImplementation() {
		public Term evaluate(Term args[]) {
		    if (args[0] instanceof MapTerm && args[1] instanceof MapTerm) {
			return ((MapTerm)args[0]).disjoint_union((MapTerm)args[1]);
		    }
		    return null;
		}
	    });
	
	spec.insertNativeImplementation("mapped-set _", new NativeImplementation() {
		public Term evaluate(Term args[]) {
		    if (args[0] instanceof MapTerm) {
			return ((MapTerm)args[0]).mapped_set();
		    }
		    return null;
		}
	    });
    }

    
    public static void register(CompiledSpecification spec) {
	new MapImplementation(spec);
    }

}
