package extensions;

import br.ufpe.abaco.compiler.*;
import br.ufpe.abaco.ParserGen.*;
import java.util.*;

public class KernelInterpreter {
  

  //OpCodes
  final static int PUSH      =  1;
  final static int TOP       =  2;
  final static int POP       =  3;
  final static int FRAME     =  4;
  final static int UNFRAME   =  5;

  final static int LRETR     =  10;
  final static int SELECT    =  11;
  final static int UNSELECT  =  12;

  final static int GRETR     =  20;
  final static int GUPDATE   =  21; 

  final static int ARGUMENTS = 30;
  final static int APPLY     = 31;
  final static int CONSTANT  = 32;

 
  class ByteCode {
    int op;
    int facet = -1;
    Term value = null;
    Operator opApply = null;

    ByteCode(int op,int facet) {
      this.op = op;
      this.facet = facet;
    }
    ByteCode(int op, Term value) {
      this.op = op;
      this.value = value;
    }
    ByteCode(int op, Operator apply) {
      this.op = op;
      this.opApply = apply;
    }

    public String toString() {
      switch (op) {
      case PUSH      : return "push " + facet;
      case TOP       : return "top " + facet; 
      case POP       : return "pop " + facet; 
      case FRAME     : return "frame " + facet; 
      case UNFRAME   : return "unframe " + facet; 
	
      case LRETR     : return "lretr " + facet; 
      case SELECT    : return "select " + facet; 
      case UNSELECT  : return "unselect " + facet; 
	
      case GRETR     : return "gretr " + facet; 
      case GUPDATE   : return "gupdate " + facet; 

      case ARGUMENTS : return "arguments " + facet; 
      case APPLY     : return "apply" + opApply; 
      case CONSTANT  : return "const " + value;
      }
      return "???";
    }
  }


  Vector localFacets = new Vector();
  int findLocal(String name) {
    int pos = localFacets.indexOf(name);
    if (pos==-1) {
      localFacets.addElement(name);
      pos = localFacets.indexOf(name);
    }
    return pos;
  }

  Vector stackFacets = new Vector();
  int findStack(String name) {
    int pos = stackFacets.indexOf(name);
    if (pos==-1) {
      stackFacets.addElement(name);
      pos = stackFacets.indexOf(name);
    }
    return pos;
  }
  

  Vector globalFacets = new Vector();
  int findGlobal(String name) {
    int pos = globalFacets.indexOf(name);
    if (pos==-1) {
      globalFacets.addElement(name);
      pos = globalFacets.indexOf(name);
    }
    return pos;
  }


  CompiledSpecification spec;
  ByteCode code[];


  KernelInterpreter(CompiledSpecification s, Term program) {
    spec = s;
    Vector vCode = new Vector();
    compile(program,vCode);
    code = (ByteCode[])vCode.toArray(new ByteCode[0]);
    
    System.out.println("Codigo compilado : ");
    for (int c=0;c<code.length;c++) {
      System.out.println("  " + c + "  " + code[c]);
    }
    System.out.println("Fim");

  }


  boolean isConstant(Term prog) {
    String progOp;
    if (prog.getId()==null)
      progOp = "";
    else 
      progOp = prog.getId().toString();
    if (progOp.equals("LOCAL RETRIEVE _")) 
      return false;
    else if (progOp.equals("TOP _")) 
      return false;
    else if (progOp.equals("POP _")) 
      return false;
    else if (progOp.equals("GLOBAL RETRIEVE _")) 
      return false;
    else {
      for (int c=0;c<prog.arity();c++)
	if (!isConstant(prog.getLeaf(c)))
	  return false;
    }
    return true;
  }

  void evaluate(Term prog, Vector resul) {
    String progOp;
    if (prog.getId()==null)
      progOp = "";
    else 
      progOp = prog.getId().toString();

    if (progOp.equals("LOCAL RETRIEVE _")) {
      int facet = findLocal(prog.getLeaf(0).toString());
      resul.addElement(new ByteCode(LRETR,facet));
    }
    else if (progOp.equals("TOP _")) {
      int facet = findStack(prog.getLeaf(0).toString());
      resul.addElement(new ByteCode(TOP,facet));
    }
    else if (progOp.equals("POP _")) {
      int facet = findStack(prog.getLeaf(0).toString());
      resul.addElement(new ByteCode(TOP,facet));
      resul.addElement(new ByteCode(POP,facet));
    }

    else if (progOp.equals("GLOBAL RETRIEVE _")) {
      int facet = findGlobal(prog.getLeaf(0).toString());
      resul.addElement(new ByteCode(GRETR,facet));
    } 
    else if (isConstant(prog)) { 
      resul.addElement(new ByteCode(CONSTANT,prog));
    } else {
      resul.addElement(new ByteCode(ARGUMENTS,prog.arity()));
      for (int c=0;c<prog.arity();c++)
	evaluate(prog.getLeaf(c),resul);
      resul.addElement(new ByteCode(APPLY,prog.getId()));
    }    
  }
  
  void compile(Term prog, Vector resul) {
    String progOp = prog.getId().toString();

    if (progOp.equals("_ AND THEN _")) {
      compile(prog.getLeaf(0),resul);
      compile(prog.getLeaf(1),resul);
    } else if (progOp.equals("COMPLETE")) {
      return ;
    } 
    else if (progOp.equals("PUSH _ _")) {
      int facet = findStack(prog.getLeaf(0).toString());
      evaluate(prog.getLeaf(1),resul);
      resul.addElement(new ByteCode(PUSH,facet));
    } else if (progOp.equals("POP _")) {
      int facet = findStack(prog.getLeaf(0).toString());
      resul.addElement(new ByteCode(POP,facet));
    } else if (progOp.equals("FRAME _ _")) {
      int facet = findStack(prog.getLeaf(0).toString());
      resul.addElement(new ByteCode(FRAME,facet));
      compile(prog.getLeaf(1),resul);
      resul.addElement(new ByteCode(UNFRAME,facet));
      
    } 
    else if (progOp.equals("SELECT _ _ _")) {
      int facet = findLocal(prog.getLeaf(0).toString());
      evaluate(prog.getLeaf(1),resul);
      resul.addElement(new ByteCode(SELECT,facet));
      compile(prog.getLeaf(2),resul);
      resul.addElement(new ByteCode(UNSELECT,facet));
    } 
    
    else if (progOp.equals("UPDATE _ _")) {
      int facet = findGlobal(prog.getLeaf(0).toString());
      evaluate(prog.getLeaf(1),resul);
      resul.addElement(new ByteCode(GUPDATE,facet));
    }
    else {
      System.out.println("Unkown Kernel Action");
    }
  }


  class Stack1 {
    Term curr;
    Stack1 next;
    Stack1(Term c,Stack1 n) { curr = c;next = n;}
  }

  class Stack2 {
    Stack1 curr;
    Stack2 next;
    Stack2(Stack1 c, Stack2 n) { curr = c;next = n;}
  }


  Term execute() {
    Term globalStorage[] = new Term[globalFacets.size()];
    Stack1 localValues[] = new Stack1[localFacets.size()];
    Stack2 stackValues[] = new Stack2[stackFacets.size()];
    int pc = 0;
    
    while (pc<code.length) {
      ByteCode curr = code[pc];
      /*switch (curr.op) {
	case 
      }*/
    }

    return null;
  }



  public static void register(final CompiledSpecification spec) {
    
    spec.insertNativeImplementation("perform _", new NativeImplementation() {
      public Term evaluate(Term args[]) {
	System.out.println("Executando Acao");
	KernelInterpreter i = new KernelInterpreter(spec,args[0]);
	return i.execute();
      }
    });
  }
 
}
